
	SECTION('Classes')
Replacer          EQUATE('<0EFh,0BFh,0BD>')  ! UTF-8 replacement character:  U+FFFD
UTF8IndexQueue    QUEUE,TYPE
!  Each record contains information about one UTF-8 character
Index                   LONG         ! index in the string where the cahracter appears
Value                   STRING(4)    ! The value of the character, which may occupy up to four positions in a Clarion string
Count                   BYTE         ! The size of the character, in bytes (1-4)
Valid                   BYTE         ! TRUE if the value at "Index" is valid UTF-8, FALSE otherwise
					END

SUTF8              CLASS,TYPE
XQ                      &UTF8IndexQueue
STR                     &STRING,PRIVATE  ! the complete UTF-8 string
sValid                  BYTE  ! TRUE if entire string contains valid UTF-8
sBytes                  LONG  ! length of string in bytes
cLEN                    LONG  ! number of characters in string
Inited                  BYTE,PRIVATE
!
Init                 PROCEDURE(*STRING S)  ! Must be called before any other method
uSlice               PROCEDURE(LONG First, LONG Last),STRING  ! Slice
uSub                 PROCEDURE(LONG First, LONG Size),STRING  ! Substring
Construct            PROCEDURE
Destruct             PROCEDURE
                   END

UTF8StringManager   CLASS,TYPE
! UTF-8 equivalents for Clarion built-ins

! First and Last are character indexes, not byte indexes
uSlice               PROCEDURE(*STRING S, LONG First, LONG Last, *STRING Out),BYTE,PROC  ! SLICE

! First is character index, Count is character count
uSub                 PROCEDURE(*STRING S, LONG First, LONG Count, *STRING Out),BYTE,PROC ! SUB

! Length is character count
uAll                 PROCEDURE(*STRING Pattern, LONG Length=255, *STRING Out),BYTE,PROC  ! ALL

! S is a single UTF-8 character, result is integer code point value
uVal                 PROCEDURE(STRING S),LONG                                            ! VAL

! c is integer code point value, result is UTF-8 character (replacement returned for invalid characters)
uChr                 PROCEDURE(ULONG c),STRING                                           ! CHR

! other UTF-8 functions not needed:
! INSTRING and INLIST return correct position for valid inputs
! UPPER and LOWER operate only on ASCII.  All other scripts are special case
! CLIP, CENTER, LEFT, RIGHT manipulate leading and trailing spaces transparently to UTF-8
! NUMERIC and DEFORMAT have no UTF-8 equvalence
                   END

    SECTION('ClassCode')
! =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
!    UTF8StringManager
UTF8StringManager.uAll  PROCEDURE(*STRING Pattern, LONG Length=255, *STRING Out)
Pat     SUTF8  
Test    SUTF8
K                           LONG
S     &STRING
S200   STRING(200)

	CODE
		Pat.Init(Pattern)
		K =  Pat.sBytes * (INT(Length/Pat.cLEN) + 1)
		IF NOT Pat.sValid 
			Out = ''
			RETURN 1
		ELSE  
			S &= NEW STRING(K)
			S = ALL(Pat.STR,K)
			Test.Init(S)
			Out = Test.uSlice(1,Length)
			RETURN 0
		END

UTF8StringManager.uSlice PROCEDURE(*STRING S, LONG First, LONG Last, *STRING Out) 
Udat     SUTF8  
	CODE
		UDat.Init(S)
		IF UDat.sValid
			Out = UDat.uSlice(First,Last)
			RETURN 0
		ELSE 
			Out = ''
			RETURN 1
		END	
		
UTF8StringManager.uSub PROCEDURE(*STRING S, LONG First, LONG Size,*STRING Out)  
Udat     SUTF8  
	CODE
		UDat.Init(S)
		IF UDat.sValid
			Out = UDat.uSub(First,Size)
			RETURN 0
		ELSE 
			Out = ''
			RETURN 1
		END	

UTF8StringManager.uVal   PROCEDURE(STRING S)
! Returns the value of a UTF-8 encoded code point (=the UTF-32 encoding)
! Applies bitwise operations per the UTF-8 encoding definition
Z                       LONG          ! result accumulator
thisbyte                BYTE,AUTO     ! processing area for current byte
SI                      STRING(1),OVER(thisbyte)  ! its string personality
I                       BYTE,AUTO     ! byte index
K                       BYTE,AUTO     ! character length
	CODE 
		ASSERT(ValidUtf8(S,1),'Input to UVal must be a valid UTF-8 character')
		ASSERT(LEN(S) <= 4,'Input to UVal was incorrect length')
		K = LEN(CLIP(S)) ! should not depend on LEN() - big hole
		LOOP I = 1 TO K
			SI = S[I]                      ! fetch the Ith byte of the character
			Z = BSHIFT(Z,6) + BAND(thisbyte,CHOOSE(I>1,3Fh,CHOOSE(K,7Fh,1Fh,0Fh,7))) ! accumulate value
		END
    RETURN Z			
		
UTF8StringManager.uChr           PROCEDURE(ULONG c)
! Computes the UTF-8 character for a legal Unicode code point.
! The algorithm encodes the number c in six-bit chunks, least significant bits first.
! The defined replacement character is returned for illegal code points.
I                       BYTE           ! byte counter
Z                       STRING(4)      ! result holder
thisbyte                BYTE,AUTO      ! current result byte construction area
S                       STRING(1),OVER(thisbyte)  ! string personality of thisbyte
	CODE
	  IF c > 10FFFFh THEN RETURN Replacer. ! Input number is too large. return the replacer.
	  IF c < 80h THEN RETURN CHR(c).     ! Input number is a US-ASCII code point, use CHR
	  LOOP                               ! process six least signicant bits of c per pass through loop
	    IF c < 20h                       ! if no more than five bits of significance remain, no more to encode
	        thisbyte = C + CHOOSE(I,0C0h,0E0h,0F0h)  ! add required number of leading 1-bits to the leading byte
	    ELSE                             ! encode a tail byte
	        thisbyte = 80h + BAND(3Fh,C) ! encode the least six bits into a tail byte
    	    I += 1                       ! count tail bytes
	    END   
	    Z = S & Z                        ! Prepend the constructed byte to the result
 	    C = BSHIFT(C,-6)                 ! shift c six bits right for next cycle (if any)
	  UNTIL thisbyte>0BFh                ! exit loop when all bytes encoded
	  IF ValidUTF8(Z) 
  		RETURN(CLIP(Z))                  ! return a valid UTF-8 character
	  ELSE
	    RETURN Replacer                  ! Replace an illegal code with the replacement character
    END
! =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
SUTF8.Construct PROCEDURE
	CODE
		Self.XQ &= NEW UTF8IndexQueue

SUTF8.Destruct PROCEDURE
	CODE
		IF Self.Inited
			DISPOSE(Self.STR)
		END  
		FREE(Self.XQ)
		DISPOSE(Self.XQ)

SUTF8.uSlice PROCEDURE(LONG First, LONG Last)
! Equivalent to Clarion slice, 
! but first and last indicate UTF-8 character positions, not byte positions
Min      LONG
Max      LONG
	CODE
		ASSERT(Self.Inited,'Programming Error - SUTF8 object not initiated before calling uSlice method')
		ASSERT(First>0 AND (Last>0), 'Programming Error - Arguments of uSlice must be greater than 0')
		ASSERT(First<=Self.cLEN AND (Last<=Self.cLEN), 'Programming Error - Arguments of uSlice must not be greater than string length')
		ASSERT(Last>=First,'Programming Error - uSlice arguments:  Last < First')
		First = CHOOSE( First <= Self.cLEN,  CHOOSE(First>0,First,1)  , Self.cLEN)
		Last = CHOOSE( Last <= Self.cLEN,  CHOOSE(Last>0,Last,1)  , Self.cLEN)
		Last = Choose(Last>=First,Last,First)
		IF Self.sValid	
			GET(Self.XQ,First)
			Min = Self.XQ.Index
			GET(Self.XQ,Last)
			Max = Self.XQ.Index + Self.XQ.Count - 1
			RETURN Self.STR[Min : Max]
		ELSE
			RETURN ''
		END	
     
SUTF8.uSub PROCEDURE(LONG First, LONG Size)
! Equivalent to Clarion SUB,
! but first indicates UTF-8 character position, not byte position
! Size indicates number of UTF-8 characters
k               LONG
	CODE
		ASSERT(Self.Inited,'Programming Error')
		IF First > 0  ! slice relative to start of string
			RETURN(SELF.uSlice(First,First+Size-1))
		ELSE  ! slice relative to end of string
			k = RECORDS(Self.XQ)+1+First
			RETURN(Self.uSlice(k+1-Size,k))
		END
		
 
SUTF8.Init          PROCEDURE(*STRING S) 

    OMIT('****')
   The purpose of this function is to map all of the UTF-8 characters in string S into SELF.XQ
   
   This function is RFC 3629 compliant in the following sense:
   
   RFC 3629 states "Implementations of the decoding algorithm MUST protect against decoding invalid sequences."
   The Unicode Standard requires decoders to "...treat any ill-formed code unit sequence as an error condition. 
   This guarantees that it will neither interpret nor emit an ill-formed code unit sequence." 

   Many UTF-8 decoders throw an exception if a string has an error in it. 
   This is impractical: being unable to work with data means you cannot even try to fix it.  
   A more useful approach used by this code is to translate the first byte to a replacement 
   character and continue parsing with the next byte.  

   This permits handling data that may have been garbled in transmission.  

    ****

Decoder  GROUP
! This table does not indicate validity - it is used to determine 
! character length, assuming valid UTF-8 characters.
x_00:7F      STRING('<1>{128}')  ! US ASCII codes: 128 values
x_80:BF      STRING('<0>{64}')   ! UTF-8 contination bytes
x_C0:C1      STRING('<255>{2}')  ! First byte of 2-byte overlong encodings
x_C2:DF      STRING('<2>{30}')   ! 2-byte character start  
x_E0:EF      STRING('<3>{16}')   ! 3-byte character start  
x_F0:F4      STRING('<4>{5}')    ! 4-byte character start  
x_F4:FF      STRING('<255>{11}') ! Invalid and restricted codes
  END 
Codes        BYTE,DIM(256),OVER(Decoder)


Testcodes   BYTE(7),DIM(100)  
valid   BYTE(1)
L       LONG
b       LONG ! byte index
XQ      &UTF8IndexQueue

U          STRING(1)          
V          BYTE,OVER(U)
cv         BYTE ! type of V (0, 1, 2, 3, 4, or 255)

  CODE
  XQ &= SELF.XQ  ! simplify coding by using a reference
  L = LEN(CLIP(S))
  Self.sBytes = L  ! string length in bytes
  SELF.STR &= NEW STRING(L)
  SELF.STR = S
  Valid = ValidUTF8(S)  ! Test entire string for validity
  Self.sValid = Valid
  LOOP b = 1 TO L
    U = S[b]    
    cv = Codes[V+1]  ! V is 0-origin, must adjust for 1-origin Clarion indexing 
    IF cv < 255
      XQ.Count = cv
      XQ.Index = b
      XQ.Value = SUB(S,b,b+cv-1)
    END
    IF cv > 0  
      IF Valid OR ValidUTF8(XQ.Value) ! Note: no test for the 255 case needed
        XQ.Valid = TRUE
      ELSE
        xQ.Valid = FALSE
        xQ.Value = Replacer & ''
        SELF.STR[b : b - 1 + LEN(CLIP(XQ.Value))] = XQ.Value  ! Replace invalid byte 
        XQ.Count = 1                    ! Set character count to 1
      END

      ADD(XQ)
    END     
  END 
  Self.cLEN = RECORDS(XQ)
  Self.Inited = TRUE

	SECTION('ValidityChecker')
ValidUtf8           PROCEDURE(*STRING S,BYTE Single=0)

   OMIT('****')
  Notes on the validity checker:

  RFC 3629 (UTF-8 standard) states "Implementations of the decoding algorithm MUST 
  protect against decoding invalid sequences ... Decoding invalid sequences may 
  have security consequences or cause other problems."

  The Unicode Standard requires decoders to:
  "...treat any ill-formed code unit sequence as an error condition.  This guarantees that it will neither 
  interpret nor emit an ill-formed code unit sequence." 

  This procedure uses a Clarion MATCH statement to validate an asserted UTF-8 string.   The match
  uses a regular expression that excludes all overlongs, surrogates, and otherwise restricted or invalid codes.
  
  The expression is based on Section 4 of RFC 3629, which shows the valid hex
  UTF-8 sequences in Augmented BNF.  The ABNF as defined in RFC 3629:
  
   UTF8-octets = *( UTF8-char )
   UTF8-char   = UTF8-1 / UTF8-2 / UTF8-3 / UTF8-4
   UTF8-1      = %x00-7F
   UTF8-2      = %xC2-DF UTF8-tail
   UTF8-3      = %xE0 %xA0-BF UTF8-tail / %xE1-EC 2( UTF8-tail ) /
                 %xED %x80-9F UTF8-tail / %xEE-EF 2( UTF8-tail )
   UTF8-4      = %xF0 %x90-BF 2( UTF8-tail ) / %xF1-F3 3( UTF8-tail ) /
                 %xF4 %x80-8F 2( UTF8-tail )
   UTF8-tail   = %x80-BF

  A Perl regular expression based on the ABNF definition can be found at:
      http://www.w3.org/International/questions/qa-forms-utf-8

  The Perl expression:  (Author: Martin Drst, W3C)      
$field =~
  m/\A(
     [\x09\x0A\x0D\x20-\x7E]            # ASCII
   | [\xC2-\xDF][\x80-\xBF]             # non-overlong 2-byte
   |  \xE0[\xA0-\xBF][\x80-\xBF]        # excluding overlongs
   | [\xE1-\xEC\xEE\xEF][\x80-\xBF]{2}  # straight 3-byte
   |  \xED[\x80-\x9F][\x80-\xBF]        # excluding surrogates
   |  \xF0[\x90-\xBF][\x80-\xBF]{2}     # planes 1-3
   | [\xF1-\xF3][\x80-\xBF]{3}          # planes 4-15
   |  \xF4[\x80-\x8F][\x80-\xBF]{2}     # plane 16
  )*\z/x;

 The Perl version above permits Perl comments and white space.  The Clarion version
 below will validate any UTF-8 text that can fit into a Clarion string, including entire
 UTF-8 files.  Special checks for white space or comments are unnecessary.
 
 Note 1: Both the expression above, and the Clarion version exclude most ASCII control codes,
 despite these being valid UTF-8.  Additional ASCII control codes may be needed in some contexts.


 End of omit  ****
! 
! We first define a set of equates comprising the fundamental building blocks of the expression,
! in the style of the ABNF form shown above:
! ---------------------------------------------
! One byte UTF-8 characters (US-ASCII):
U1  EQUATE('[<09h,0Ah,0Dh><20h>-<7Eh>]') 
! Note that we omitted all ASCII control codes except tab, line feed, and carriage return
! This would be overly strict for a general procedure, but this code is intended for Clarion STRING use.

! Tail byte that terminates all multibyte UTF-8 characters:
Tail     EQUATE('[<80h>-<0BFh>]')

! Two byte UTF-8 characters:
!    (Two byte characters beginning C0-C1 are omitted because they are "overlongs")
U2        EQUATE('[<0C2h>-<0DFh>]' & Tail)              

! Three byte UTF-8 character forms:
U3_1      EQUATE('<0E0h>[<0A0h>-<0BFh>]' & Tail)
U3_2      EQUATE('[<0E1h>-<0ECh><0EEh,0EFh>]' & Tail & Tail)
U3_3      EQUATE('<0EDh>[<080h>-<09Fh>]' & Tail)       

! Four byte UTF-8 character forms:
U4_1      EQUATE('<0F0h>[<090h>-<0BFh>]' & Tail & Tail) 
U4_2      EQUATE('[<0F1h>-<0F3h>]' & Tail & Tail & Tail)
U4_3      EQUATE('<0F4h>[<080h>-<08Fh>]' & Tail & Tail) 

! Putting it all together:
! Clarion Regular Expression for UTF-8, comprising all valid character forms:
! ---------------------------------------------
RegValid      STRING('^{{' & U1 & '|' & |
                     U2 & '|' & |
                     U3_1 & '|' & |
                     U3_2 & '|' & |
                     U3_3 & '|' & |
                     U4_1 & '|' & |
                     U4_2 & '|' & |
                     U4_3 & '}*$')
	CODE
      IF Single THEN RegValid[LEN(RegValid)-1]='?'.
      RETURN(MATCH(S,RegValid,Match:Regular))  ! That's all there is, folks!

